/*
 * Copyright (C) 2005 Cesar Kallas
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * E-mail: cesarkallas@gmx.net
 * http://cesarkallas.soulivre.org
 *
 */
 
package estella.cliente;
import  estella.lib.*;
import  estella.gui.*;

import java.io.File;
import java.io.FileReader;
import java.io.BufferedReader;
import java.util.StringTokenizer;
import java.util.Vector;

/**
* Funcoes do cliente
* @author Cesar Kallas
* @since 21/11/2005
*/
public class Cliente 
{
	public javax.swing.JFrame jFrame;
	private Constantes CON;
	private Log LOG;
  private Tempo TEMPO;
  private StringTokenizer preServents;
	private Payload payload;
	private Descritor descritor;
	private Thread t_recebeDescritores;

	/**
	* Tenta conecta aos servents declarados na lista de servents pre-definidas em um arquivo
	* @return true ou false
	*/
  public boolean conecta(String host, int porta)
  {
		System.out.println("Conectar em : " + host + " porta: " + porta);
		try {
			// Tenta conectar, verifica resposta do servidor e se ok, adiciona a lista de conexoes
			Conexao conexao = new Conexao(host, porta);
			if(conexao.inicia()) {
				conexao.setTimeout(CON.SOCKET_TIMEOUT);
				conexao.envia(CON.GNUTELLA);
				if(conexao.recebe().indexOf(CON.GNUTELLA_OK) != -1) {
					conexao.setTimeout(0);
					//LOG.info("Conectado ao servent " + conexao.status());
					synchronized(Dados.onlineServents) {
						Dados.onlineServents.adiciona(conexao);
					}
					synchronized(JanelaPrincipal.labelServentsOnline) {
						JanelaPrincipal.labelServentsOnline.setText(Dados.onlineServents.tamanho() + "");
					}
					synchronized(JanelaPrincipal.labelOnline) {
						JanelaPrincipal.labelOnline.setText("sim");
					}
					synchronized(JanelaPrincipal.linhaTabelaServentsOnline) {
						Vector serv = new Vector();
						serv.addElement(host);
						serv.addElement(new String(porta + ""));
						JanelaPrincipal.linhaTabelaServentsOnline.addElement(serv);
						synchronized(JanelaPrincipal.tabelaServentsOnline) {
							JanelaPrincipal.tabelaServentsOnline.addNotify();
						}
					}
					t_recebeDescritores = new Thread(new recebeDescritores(conexao));
					t_recebeDescritores.start();
					return true;
				}
			}
			return false;
		}
		catch(Exception FalhaConectar) {
			LOG.falha("Nao consegui conectar no " + host + "@" + porta + " ->" + FalhaConectar.getMessage());
			return false;
		}
	}
	
	public boolean ping(Conexao conexao)
	{
		payload = new Payload();
		descritor = new Descritor(CON.PING, CON.TTL, 0, payload.ping());
		if(conexao.envia(new String(descritor.retorna())) == false)
			return false;
		return true;
	}
	
	public boolean query(Conexao conexao, int minVelocidade, String busca)
	{
		payload = new Payload();
		descritor = new Descritor(CON.QUERY, CON.TTL, 0, payload.query(minVelocidade, busca));
		if(conexao.envia(new String(descritor.retorna())) == false)
			return false;
		synchronized(Dados.queryEnviados) {
			Dados.queryEnviados.adiciona(descritor);
			LOG.info("Query adcionada na lista! " + descritor.stringId());
		}
		return true;
	}

	/**
	* Retorna uma lista de servents pre-definidas em um arquivo
	* @param caminho_servents String com o caminho para o arquivo com os servents. Arquivo definido no formato --  'servent:porta' --
	* @return String com todos os servents e portas encontrados no arquivo
	*/
	public boolean carregaServents(String caminho_servents)
	{
		BufferedReader lista_servents;
		String servent, servents;
		int i=0;
		
		try {
			lista_servents = new BufferedReader(new FileReader(new File(caminho_servents)));
			servent = new String();
			
			while(servent != null) {
				servent = lista_servents.readLine();
				if(servent == null)
					return true;
				Vector tabelaServent = new Vector();
				StringTokenizer st = new StringTokenizer(servent, ":");
				tabelaServent.addElement(st.nextToken());
				tabelaServent.addElement(st.nextToken());
				synchronized(JanelaPrincipal.tabelaServents) {
					synchronized(JanelaPrincipal.linhaTabelaServents) {
						JanelaPrincipal.linhaTabelaServents.addElement(tabelaServent);
					}
					JanelaPrincipal.tabelaServents.addNotify();
				}
					//servents += servent + CON.SEPARADOR;
			}
			lista_servents.close();
			return true;
		}
		catch(Exception LerServents) {
			LOG.falha(LerServents.getMessage());
			return false;
		}
	}
}

/**
* Thread para tratar requisicao recebida pelo servidor
* @param sck_cliente Socket com a conexao do cliente
* @since 23/11/2005
* @author Cesar Kallas - cesarkallas at gmx . net
*/
class recebeDescritores implements Runnable
{
  private Conexao conexao;
  private Log LOG;
  private Constantes CON;
	private Descritor descritor;
	private String mensagem;
	private Dados dados;
	private Thread t_trataDescritor;
		
  public recebeDescritores(Conexao cliente)
  {
    LOG = new Log("Cliente");
    CON = new Constantes();
		conexao = cliente;
  }

  public void run()
  {
		LOG.info("Escutando descritores de " + conexao.status());
		// Tratar descritores que cheguem ao cliente
		while(conexao.status() != null) {
			try {
				mensagem = new String(conexao.recebe());
			}
			catch(Exception ReceberDescritor)
			{
				LOG.falha(ReceberDescritor.getMessage());
				return;
			}
			descritor = new Descritor(mensagem);
			t_trataDescritor = new Thread(new trataDescritor(descritor, conexao));
			t_trataDescritor.start();
		}
		synchronized(Dados.onlineServents) {
			Dados.onlineServents.remove(conexao);
		}
		LOG.info("Cliente desconectado");
		return;
	}
}

class trataDescritor implements Runnable
{
	private Descritor descritor;
	private Log LOG;
	private Constantes CON;
	private Conexao conexao;
	private Payload payload;
		
	public trataDescritor(Descritor descritor, Conexao conexao)
	{
		this.descritor = descritor;
		this.conexao = conexao;
		this.CON = new Constantes();
		this.LOG = new Log("Cliente");
	}
	
	public void run()
	{
		// Tamanho minimo de descritor 20 (id + ttl + hops + payload length)
		if(descritor.retorna().length < 20)
			return;
		
		switch(descritor.tipo())
		{
			case (byte)00000000: break;
			case (byte)00000001: { pong(); break; }
			case (byte)10000001: { queryhits(); break; }
			default: LOG.info("ho ho...o que eh isso?"); return;
		}
	}
	
	private void pong()
	{		
		LOG.info("pong recebido id(" + descritor.stringId() + ") ttl(" + descritor.ttl() + ") hops(" + descritor.hops() +  ") <- " + conexao.status());
		Tempo tempo = new Tempo();
		Payload payload = new Payload(CON.PONG, descritor.payload());
		Servent servent = new Servent(payload.pongIp(), payload.pongPorta(), payload.pongNumArqComp(), payload.pongTamArqComp(), tempo.data()+"  "+tempo.hora());
		
		synchronized(Dados.servents) {
			Dados.servents.adiciona(servent);
		}
		LOG.info("servent cadastrado IP: " + payload.pongIp() + " porta: " + payload.pongPorta() + " numArqComp: " + payload.pongNumArqComp() + " tamArqComp: " + payload.pongTamArqComp());
	}
	
	private void queryhits()
	{
		LOG.info("queryhits recebido id(" + descritor.stringId() + ") ttl(" + descritor.ttl() + ") hops(" + descritor.hops() +  ") <- " + conexao.status());
		Payload payload = new Payload(CON.QUERYHITS, descritor.payload());
		LOG.info("resultado da pesquisa - hits: " + payload.queryhitsNum() + " ip: " + payload.queryhitsIp() + " porta: " + payload.queryhitsPorta());

		byte[] resultset = payload.queryhitsResultset();
		
		// Verifica arquivos recebidos
		synchronized(Dados.queryEnviados) {
			if(Dados.queryEnviados.existe(descritor.stringId()))
				LOG.info(new String(payload.retornaHit(payload.queryhitsNum(), resultset)));
			else
				LOG.atencao("Query recebido, mas nao localizado na lista de enviados");
		}
	}
}
